@@ -5,7 +5,6 @@ import pathlib
 from pathlib import Path
 import pandas as pd
 import shutil
-from urllib.parse import urlparse
 
 import jsonlines
 from unittest.mock import patch, call, ANY
@@ -13,52 +12,21 @@ import pytest
 
 import quilt3
 from quilt3 import Package
-from quilt3.util import QuiltException, validate_package_name, parse_file_url, fix_url
+from quilt3.util import PhysicalKey, QuiltException, validate_package_name, fix_url
 
 from ..utils import QuiltTestCase
 
 
 DATA_DIR = Path(__file__).parent / 'data'
-SERIALIZATION_DIR = Path(__file__).parent / 'serialization_dir'
 LOCAL_MANIFEST = DATA_DIR / 'local_manifest.jsonl'
 REMOTE_MANIFEST = DATA_DIR / 'quilt_manifest.jsonl'
 
+SERIALIZATION_DIR = Path('serialization_dir')
+
 LOCAL_REGISTRY = Path('local_registry')  # Set by QuiltTestCase
 
 
-def mock_make_api_call(self, operation_name, kwarg):
-    """ Mock boto3's AWS API Calls for testing. """
-    if operation_name == 'GetObject':
-        parsed_response = {'Body': BytesIO(b'foo')}
-        return parsed_response
-    if operation_name == 'ListObjectsV2':
-        parsed_response = {'CommonPrefixes': ['foo']}
-        return parsed_response
-    if operation_name == 'HeadObject':
-        # TODO: mock this somehow
-        parsed_response = {
-            'ContentLength': 0
-        }
-        return parsed_response
-    raise NotImplementedError(operation_name)
-
-
-
 class PackageTest(QuiltTestCase):
-    def setUp(self):
-        super().setUp()
-        self.file_sweeper_path_list = []
-
-    def tearDown(self):
-        super().tearDown()
-        if len(self.file_sweeper_path_list) > 0:
-            for fpath in self.file_sweeper_path_list:
-                if os.path.exists(fpath):
-                    try:
-                        os.remove(fpath)
-                    except Exception as e:
-                        print("Error when removing file", fpath, str(e))
-
     def test_build(self):
         """Verify that build dumps the manifest to appdirs directory."""
         new_pkg = Package()
@@ -77,7 +45,7 @@ class PackageTest(QuiltTestCase):
         out_path = LOCAL_REGISTRY / ".quilt/packages" / top_hash
         with open(out_path) as fd:
             pkg = Package.load(fd)
-            assert test_file.resolve().as_uri() == pkg['foo'].physical_keys[0]
+            assert PhysicalKey.from_path(test_file) == pkg['foo'].physical_key
 
         # Verify latest points to the new location.
         named_pointer_path = LOCAL_REGISTRY / ".quilt/named_packages/Quilt/Test/latest"
@@ -91,7 +59,7 @@ class PackageTest(QuiltTestCase):
         out_path = LOCAL_REGISTRY / ".quilt/packages" / top_hash
         with open(out_path) as fd:
             pkg = Package.load(fd)
-            assert test_file.resolve().as_uri() == pkg['bar'].physical_keys[0]
+            assert PhysicalKey.from_path(test_file) == pkg['bar'].physical_key
 
     def test_default_registry(self):
         new_pkg = Package()
@@ -110,7 +78,7 @@ class PackageTest(QuiltTestCase):
         out_path = LOCAL_REGISTRY / ".quilt/packages" / top_hash
         with open(out_path) as fd:
             pkg = Package.load(fd)
-            assert test_file.resolve().as_uri() == pkg['foo'].physical_keys[0]
+            assert PhysicalKey.from_path(test_file) == pkg['foo'].physical_key
 
         # Verify latest points to the new location.
         named_pointer_path = LOCAL_REGISTRY / ".quilt/named_packages/Quilt/Test/latest"
@@ -124,9 +92,10 @@ class PackageTest(QuiltTestCase):
         out_path = LOCAL_REGISTRY / ".quilt/packages" / top_hash
         with open(out_path) as fd:
             pkg = Package.load(fd)
-            assert test_file.resolve().as_uri() == pkg['bar'].physical_keys[0]
+            assert PhysicalKey.from_path(test_file) == pkg['bar'].physical_key
 
     @patch('quilt3.Package._browse', lambda name, registry, top_hash: Package())
+    @patch('quilt3.Package._shorten_tophash', lambda package_name, registry, top_hash: "7a67ff4")
     def test_default_install_location(self):
         """Verify that pushes to the default local install location work as expected"""
         with patch('quilt3.Package._materialize') as materialize_mock:
@@ -134,7 +103,7 @@ class PackageTest(QuiltTestCase):
             Package.install(pkg_name, registry='s3://my-test-bucket')
 
             materialize_mock.assert_called_once_with(
-                quilt3.util.get_install_location().rstrip('/') + '/' + pkg_name,
+                PhysicalKey.from_url(quilt3.util.get_install_location()).join(pkg_name),
             )
 
     def test_read_manifest(self):
@@ -154,8 +123,12 @@ class PackageTest(QuiltTestCase):
         with open(out_path) as fd:
             written_set = list(jsonlines.Reader(fd))
         assert len(original_set) == len(written_set)
-        assert sorted(original_set, key=lambda k: k.get('logical_key', 'manifest')) \
-            == sorted(written_set, key=lambda k: k.get('logical_key', 'manifest'))
+
+        if os.name != 'nt':
+            # TODO: LOCAL_MANIFEST contains paths like file:///foo -
+            # but they're not valid absolute paths on Windows. What do we do?
+            assert sorted(original_set, key=lambda k: k.get('logical_key', 'manifest')) \
+                == sorted(written_set, key=lambda k: k.get('logical_key', 'manifest'))
 
     def test_remote_browse(self):
         """ Verify loading manifest from s3 """
@@ -285,29 +258,6 @@ class PackageTest(QuiltTestCase):
             Package.browse('Quilt/test', top_hash='123456', registry=registry)
 
 
-    def test_remote_install(self):
-        """Verify that installing from a local package works as expected."""
-        remote_registry = Path('.').resolve().as_uri()
-        quilt3.config(
-            default_local_registry=remote_registry,
-            default_remote_registry=remote_registry
-        )
-        with patch('quilt3.Package.push') as push_mock:
-            pkg = Package()
-            pkg.build('Quilt/nice-name')
-
-            with patch('quilt3.Package._materialize') as materialize_mock, \
-                patch('quilt3.Package._build') as build_mock:
-                materialize_mock.return_value = pkg
-                dest_registry = quilt3.util.get_from_config('default_local_registry')
-
-                quilt3.Package.install('Quilt/nice-name', dest='./')
-
-                materialize_mock.assert_called_once_with(fix_url('./'))
-                build_mock.assert_called_once_with(
-                    'Quilt/nice-name', message=None, registry=dest_registry
-                )
-
     def test_install_restrictions(self):
         """Verify that install can only operate remote -> local."""
         # disallow installs which send package data to a remote registry
@@ -339,10 +289,10 @@ class PackageTest(QuiltTestCase):
             'fetch wrote {} files; expected: {}'.format(file_count, expected)
 
         # test that package re-rooting works as expected
-        out_dir_abs_path = pathlib.Path(out_dir).resolve().as_uri()
-        assert all(
-            entry.physical_keys[0].startswith(out_dir_abs_path) for _, entry in new_package_.walk()
-        )
+        out_dir_abs_path = pathlib.Path(out_dir).resolve()
+        for _, entry in new_package_.walk():
+            # relative_to will raise an exception if the first path is not inside the second path.
+            pathlib.Path(entry.physical_key.path).relative_to(out_dir_abs_path)
 
     def test_package_fetch_default_dest(self):
         """Verify fetching a package to the default local destination."""
@@ -375,8 +325,7 @@ class PackageTest(QuiltTestCase):
         # The key gets re-rooted correctly.
         pkg = quilt3.Package().set('foo', DATA_DIR / 'foo.txt')
         new_pkg_entry = pkg['foo'].fetch('bar.txt')
-        out_abs_path = pathlib.Path("bar.txt").resolve().as_uri()
-        assert new_pkg_entry.physical_keys[0] == out_abs_path
+        assert new_pkg_entry.physical_key == PhysicalKey.from_path('bar.txt')
 
     def test_fetch_default_dest(tmpdir):
         """Verify fetching a package entry to a default destination."""
@@ -384,9 +333,13 @@ class PackageTest(QuiltTestCase):
             (Package()
              .set('foo', os.path.join(os.path.dirname(__file__), 'data', 'foo.txt'))['foo']
              .fetch())
-            filepath = fix_url(os.path.join(os.path.dirname(__file__), 'data', 'foo.txt'))
-            copy_mock.assert_called_once_with(filepath, ANY)
+            filepath = os.path.join(os.path.dirname(__file__), 'data', 'foo.txt')
+            copy_mock.assert_called_once_with(
+                PhysicalKey.from_path(filepath),
+                PhysicalKey.from_path('foo.txt')
+            )
 
+    @patch('quilt3.Package._shorten_tophash', lambda package_name, registry, top_hash: "7a67ff4")
     def test_load_into_quilt(self):
         """ Verify loading local manifest and data into S3. """
         top_hash1 = 'abbf5f171cf20bfb2313ecd8684546958cd72ac4f3ec635e4510d9c771168226'
@@ -560,21 +513,21 @@ class PackageTest(QuiltTestCase):
 
         pkg = pkg.set_dir("/", ".", meta="test_meta")
 
-        assert pathlib.Path('foo').resolve().as_uri() == pkg['foo'].physical_keys[0]
-        assert pathlib.Path('bar').resolve().as_uri() == pkg['bar'].physical_keys[0]
-        assert (bazdir / 'baz').resolve().as_uri() == pkg['foo_dir/baz_dir/baz'].physical_keys[0]
-        assert (foodir / 'bar').resolve().as_uri() == pkg['foo_dir/bar'].physical_keys[0]
+        assert PhysicalKey.from_path('foo') == pkg['foo'].physical_key
+        assert PhysicalKey.from_path('bar') == pkg['bar'].physical_key
+        assert PhysicalKey.from_path(bazdir / 'baz') == pkg['foo_dir/baz_dir/baz'].physical_key
+        assert PhysicalKey.from_path(foodir / 'bar') == pkg['foo_dir/bar'].physical_key
         assert pkg.meta == "test_meta"
 
         pkg = Package()
         pkg = pkg.set_dir('/','foo_dir/baz_dir/')
         # todo nested at set_dir site or relative to set_dir path.
-        assert (bazdir / 'baz').resolve().as_uri() == pkg['baz'].physical_keys[0]
+        assert PhysicalKey.from_path(bazdir / 'baz') == pkg['baz'].physical_key
 
         pkg = Package()
         pkg = pkg.set_dir('my_keys', 'foo_dir/baz_dir/')
         # todo nested at set_dir site or relative to set_dir path.
-        assert (bazdir / 'baz').resolve().as_uri() == pkg['my_keys/baz'].physical_keys[0]
+        assert PhysicalKey.from_path(bazdir / 'baz') == pkg['my_keys/baz'].physical_key
 
         # Verify ignoring files in the presence of a dot-quiltignore
         with open('.quiltignore', 'w') as fd:
@@ -603,15 +556,15 @@ class PackageTest(QuiltTestCase):
 
         pkg = pkg.set_dir("new_dir", ".", meta="new_test_meta")
 
-        assert pathlib.Path('foo').resolve().as_uri() == pkg['new_dir/foo'].physical_keys[0]
-        assert pathlib.Path('bar').resolve().as_uri() == pkg['new_dir/bar'].physical_keys[0]
+        assert PhysicalKey.from_path('foo') == pkg['new_dir/foo'].physical_key
+        assert PhysicalKey.from_path('bar') == pkg['new_dir/bar'].physical_key
         assert pkg['new_dir'].meta == "new_test_meta"
 
         # verify set_dir logical key shortcut
         pkg = Package()
         pkg.set_dir("/")
-        assert pathlib.Path('foo').resolve().as_uri() == pkg['foo'].physical_keys[0]
-        assert pathlib.Path('bar').resolve().as_uri() == pkg['bar'].physical_keys[0]
+        assert PhysicalKey.from_path('foo') == pkg['foo'].physical_key
+        assert PhysicalKey.from_path('bar') == pkg['bar'].physical_key
 
 
     def test_s3_set_dir(self):
@@ -627,8 +580,8 @@ class PackageTest(QuiltTestCase):
 
             pkg.set_dir('', 's3://bucket/foo/', meta='test_meta')
 
-            assert pkg['a.txt'].physical_keys[0] == 's3://bucket/foo/a.txt?versionId=xyz'
-            assert pkg['x']['y.txt'].physical_keys[0] == 's3://bucket/foo/x/y.txt'
+            assert pkg['a.txt'].get() == 's3://bucket/foo/a.txt?versionId=xyz'
+            assert pkg['x']['y.txt'].get() == 's3://bucket/foo/x/y.txt?versionId=null'
             assert pkg.meta == "test_meta"
             assert pkg['x']['y.txt'].size == 10  # GH368
 
@@ -638,8 +591,8 @@ class PackageTest(QuiltTestCase):
 
             pkg.set_dir('bar', 's3://bucket/foo')
 
-            assert pkg['bar']['a.txt'].physical_keys[0] == 's3://bucket/foo/a.txt?versionId=xyz'
-            assert pkg['bar']['x']['y.txt'].physical_keys[0] == 's3://bucket/foo/x/y.txt'
+            assert pkg['bar']['a.txt'].get() == 's3://bucket/foo/a.txt?versionId=xyz'
+            assert pkg['bar']['x']['y.txt'].get() == 's3://bucket/foo/x/y.txt?versionId=null'
             assert pkg['bar']['a.txt'].size == 10 # GH368
 
             list_object_versions_mock.assert_called_with('bucket', 'foo/')
@@ -700,17 +653,17 @@ class PackageTest(QuiltTestCase):
         pkg['bar'].meta['target'] = 'unicode'
 
         # Build a dummy file to add to the map.
-        with open('bar.txt', "w") as fd:
-            fd.write('test_file_content_string')
-            test_file = Path(fd.name)
+        test_file = Path('bar.txt')
+        test_file.write_text('test_file_content_string')
         pkg['bar'].set('bar.txt')
 
-        assert test_file.resolve().as_uri() == pkg['bar'].physical_keys[0]
+        assert PhysicalKey.from_path(test_file) == pkg['bar'].physical_key
 
         # Test shortcut codepath
         pkg = Package().set('bar.txt')
-        assert test_file.resolve().as_uri() == pkg['bar.txt'].physical_keys[0]
+        assert PhysicalKey.from_path(test_file) == pkg['bar.txt'].physical_key
 
+    @patch('quilt3.Package._shorten_tophash', lambda package_name, registry, top_hash: "7a67ff4")
     def test_set_package_entry_as_object(self):
         pkg = Package()
         nasty_string = 'a,"\tb'
@@ -732,30 +685,27 @@ class PackageTest(QuiltTestCase):
         pkg.set("mydataframe6.tsv", df, meta={'user_meta': 'blah6'})
 
         for lk, entry in pkg.walk():
-            file_path = parse_file_url(urlparse(entry.get()))
+            file_path = entry.physical_key.path
             assert pathlib.Path(file_path).exists(), "The serialization files should exist"
 
-            self.file_sweeper_path_list.append(file_path)  # Make sure files get deleted even if test fails
-
         pkg._fix_sha256()
         for lk, entry in pkg.walk():
             assert df.equals(entry.deserialize()), "The deserialized PackageEntry should be equal to the object that " \
                                                    "was serialized"
 
         # Test that push cleans up the temporary files, if and only if the serialization_location was not set
-        with patch('botocore.client.BaseClient._make_api_call', new=mock_make_api_call), \
-            patch('quilt3.Package._materialize') as materialize_mock, \
+        with patch('quilt3.Package._materialize') as materialize_mock, \
             patch('quilt3.Package._build') as build_mock:
             materialize_mock.return_value = pkg
 
             pkg.push('Quilt/test_pkg_name', 's3://test-bucket')
 
         for lk in ["mydataframe1.parquet", "mydataframe2.csv", "mydataframe3.tsv"]:
-            file_path = parse_file_url(urlparse(pkg.get(lk)))
+            file_path = pkg[lk].physical_key.path
             assert pathlib.Path(file_path).exists(), "These files should not have been deleted during push()"
 
         for lk in ["mydataframe4.parquet", "mydataframe5.csv", "mydataframe6.tsv"]:
-            file_path = parse_file_url(urlparse(pkg.get(lk)))
+            file_path = pkg[lk].physical_key.path
             assert not pathlib.Path(file_path).exists(), "These temp files should have been deleted during push()"
 
 
@@ -1015,11 +965,10 @@ class PackageTest(QuiltTestCase):
         with pytest.raises(QuiltException):
             p.push('Quilt/Test', 's3://test-bucket', dest='s3://other-test-bucket')
 
-
+    @patch('quilt3.Package._shorten_tophash', lambda package_name, registry, top_hash: "7a67ff4")
     def test_commit_message_on_push(self):
         """ Verify commit messages populate correctly on push."""
-        with patch('botocore.client.BaseClient._make_api_call', new=mock_make_api_call), \
-            patch('quilt3.Package._materialize') as materialize_mock, \
+        with patch('quilt3.Package._materialize') as materialize_mock, \
             patch('quilt3.Package._build') as build_mock:
             with open(REMOTE_MANIFEST) as fd:
                 pkg = Package.load(fd)
@@ -1190,6 +1139,7 @@ class PackageTest(QuiltTestCase):
         with pytest.raises(QuiltException):
             pkg.set('s3://foo/..', LOCAL_MANIFEST)
 
+    @patch('quilt3.Package._shorten_tophash', lambda package_name, registry, top_hash: "7a67ff4")
     def test_install(self):
         # Manifest
 
@@ -1290,7 +1240,7 @@ class PackageTest(QuiltTestCase):
 
         # ...but moving it back fixes it.
         pathlib.Path('foo2').rename(local_path)
-        assert p['foo'].get_cached_path() == str(local_path)
+        assert pathlib.Path(p['foo'].get_cached_path()) == local_path
 
         # Check that changing the contents invalidates the cache.
         local_path.write_text('omg')
@@ -1347,7 +1297,7 @@ class PackageTest(QuiltTestCase):
         with self.assertRaises(QuiltException):
             Package.rollback('quilt/blah', LOCAL_REGISTRY, good_hash)
 
-
+    @patch('quilt3.Package._shorten_tophash', lambda package_name, registry, top_hash: "7a67ff4")
     def test_verify(self):
         pkg = Package()
 
